package xyz.lychee.lagfixer.libs.adventure.text;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import java.util.Objects;
import java.util.Set;
import java.util.function.Consumer;
import java.util.function.Function;
import xyz.lychee.lagfixer.libs.adventure.key.Key;
import xyz.lychee.lagfixer.libs.adventure.text.event.ClickEvent;
import xyz.lychee.lagfixer.libs.adventure.text.event.HoverEventSource;
import xyz.lychee.lagfixer.libs.adventure.text.format.Style;
import xyz.lychee.lagfixer.libs.adventure.text.format.TextColor;
import xyz.lychee.lagfixer.libs.adventure.text.format.TextDecoration;
import xyz.lychee.lagfixer.libs.annotations.NotNull;
import xyz.lychee.lagfixer.libs.annotations.Nullable;

abstract class AbstractComponentBuilder<C extends BuildableComponent<C, B>, B extends ComponentBuilder<C, B>> implements ComponentBuilder<C, B> {
   protected List<Component> children = Collections.emptyList();
   private @Nullable Style style;
   private Style.@Nullable Builder styleBuilder;

   protected AbstractComponentBuilder() {
   }

   protected AbstractComponentBuilder(final @NotNull C component) {
      List<Component> children = component.children();
      if (!children.isEmpty()) {
         this.children = new ArrayList(children);
      }

      if (component.hasStyling()) {
         this.style = component.style();
      }

   }

   public @NotNull B append(final @NotNull Component component) {
      if (component == Component.empty()) {
         return this;
      } else {
         this.prepareChildren();
         this.children.add((Component)Objects.requireNonNull(component, "component"));
         return this;
      }
   }

   public @NotNull B append(final @NotNull Component... components) {
      return this.append((ComponentLike[])components);
   }

   public @NotNull B append(final @NotNull ComponentLike... components) {
      Objects.requireNonNull(components, "components");
      boolean prepared = false;
      int i = 0;

      for(int length = components.length; i < length; ++i) {
         Component component = ((ComponentLike)Objects.requireNonNull(components[i], "components[?]")).asComponent();
         if (component != Component.empty()) {
            if (!prepared) {
               this.prepareChildren();
               prepared = true;
            }

            this.children.add((Component)Objects.requireNonNull(component, "components[?]"));
         }
      }

      return this;
   }

   public @NotNull B append(final @NotNull Iterable<? extends ComponentLike> components) {
      Objects.requireNonNull(components, "components");
      boolean prepared = false;
      Iterator var3 = components.iterator();

      while(var3.hasNext()) {
         ComponentLike like = (ComponentLike)var3.next();
         Component component = ((ComponentLike)Objects.requireNonNull(like, "components[?]")).asComponent();
         if (component != Component.empty()) {
            if (!prepared) {
               this.prepareChildren();
               prepared = true;
            }

            this.children.add((Component)Objects.requireNonNull(component, "components[?]"));
         }
      }

      return this;
   }

   private void prepareChildren() {
      if (this.children == Collections.emptyList()) {
         this.children = new ArrayList();
      }

   }

   public @NotNull B applyDeep(final @NotNull Consumer<? super ComponentBuilder<?, ?>> consumer) {
      this.apply(consumer);
      if (this.children == Collections.emptyList()) {
         return this;
      } else {
         ListIterator<Component> it = this.children.listIterator();

         while(it.hasNext()) {
            Component child = (Component)it.next();
            if (child instanceof BuildableComponent) {
               ComponentBuilder<?, ?> childBuilder = ((BuildableComponent)child).toBuilder();
               childBuilder.applyDeep(consumer);
               it.set(childBuilder.build());
            }
         }

         return this;
      }
   }

   public @NotNull B mapChildren(final @NotNull Function<BuildableComponent<?, ?>, ? extends BuildableComponent<?, ?>> function) {
      if (this.children == Collections.emptyList()) {
         return this;
      } else {
         ListIterator<Component> it = this.children.listIterator();

         while(it.hasNext()) {
            Component child = (Component)it.next();
            if (child instanceof BuildableComponent) {
               BuildableComponent<?, ?> mappedChild = (BuildableComponent)Objects.requireNonNull((BuildableComponent)function.apply((BuildableComponent)child), "mappedChild");
               if (child != mappedChild) {
                  it.set(mappedChild);
               }
            }
         }

         return this;
      }
   }

   public @NotNull B mapChildrenDeep(final @NotNull Function<BuildableComponent<?, ?>, ? extends BuildableComponent<?, ?>> function) {
      if (this.children == Collections.emptyList()) {
         return this;
      } else {
         ListIterator<Component> it = this.children.listIterator();

         while(it.hasNext()) {
            Component child = (Component)it.next();
            if (child instanceof BuildableComponent) {
               BuildableComponent<?, ?> mappedChild = (BuildableComponent)Objects.requireNonNull((BuildableComponent)function.apply((BuildableComponent)child), "mappedChild");
               if (mappedChild.children().isEmpty()) {
                  if (child != mappedChild) {
                     it.set(mappedChild);
                  }
               } else {
                  ComponentBuilder<?, ?> builder = mappedChild.toBuilder();
                  builder.mapChildrenDeep(function);
                  it.set(builder.build());
               }
            }
         }

         return this;
      }
   }

   public @NotNull List<Component> children() {
      return Collections.unmodifiableList(this.children);
   }

   public @NotNull B style(final @NotNull Style style) {
      this.style = style;
      this.styleBuilder = null;
      return this;
   }

   public @NotNull B style(final @NotNull Consumer<Style.Builder> consumer) {
      consumer.accept(this.styleBuilder());
      return this;
   }

   public @NotNull B font(final @Nullable Key font) {
      this.styleBuilder().font(font);
      return this;
   }

   public @NotNull B color(final @Nullable TextColor color) {
      this.styleBuilder().color(color);
      return this;
   }

   public @NotNull B colorIfAbsent(final @Nullable TextColor color) {
      this.styleBuilder().colorIfAbsent(color);
      return this;
   }

   public @NotNull B decoration(final @NotNull TextDecoration decoration, final TextDecoration.@NotNull State state) {
      this.styleBuilder().decoration(decoration, state);
      return this;
   }

   public @NotNull B decorationIfAbsent(final @NotNull TextDecoration decoration, final TextDecoration.@NotNull State state) {
      this.styleBuilder().decorationIfAbsent(decoration, state);
      return this;
   }

   public @NotNull B clickEvent(final @Nullable ClickEvent event) {
      this.styleBuilder().clickEvent(event);
      return this;
   }

   public @NotNull B hoverEvent(final @Nullable HoverEventSource<?> source) {
      this.styleBuilder().hoverEvent(source);
      return this;
   }

   public @NotNull B insertion(final @Nullable String insertion) {
      this.styleBuilder().insertion(insertion);
      return this;
   }

   public @NotNull B mergeStyle(final @NotNull Component that, final @NotNull Set<Style.Merge> merges) {
      this.styleBuilder().merge(((Component)Objects.requireNonNull(that, "component")).style(), merges);
      return this;
   }

   public @NotNull B resetStyle() {
      this.style = null;
      this.styleBuilder = null;
      return this;
   }

   private Style.@NotNull Builder styleBuilder() {
      if (this.styleBuilder == null) {
         if (this.style != null) {
            this.styleBuilder = this.style.toBuilder();
            this.style = null;
         } else {
            this.styleBuilder = Style.style();
         }
      }

      return this.styleBuilder;
   }

   protected final boolean hasStyle() {
      return this.styleBuilder != null || this.style != null;
   }

   protected @NotNull Style buildStyle() {
      if (this.styleBuilder != null) {
         return this.styleBuilder.build();
      } else {
         return this.style != null ? this.style : Style.empty();
      }
   }
}
